
import bpy
import math
import mathutils
import os
import re

from . import curve_module
from . import mesh_setting_module
from . import modifier_module

import numpy as np

from os import path
from math import sin, cos, sqrt, atan2
from mathutils import Vector, Matrix

from bpy.props import *

dirname = path.dirname(__file__)
addon_name = __package__

class POLEMAKEROBJ_OT_CatenaryConnect(bpy.types.Operator, curve_module.CurveModule, modifier_module.ModifierModule, mesh_setting_module.MeshSettingModule):
    
    bl_idname = "polemaker.catenary_connect"
    bl_label = "Catenary Connect"
    bl_description = "Description"
    bl_options = {'REGISTER', 'UNDO'}

    action : EnumProperty(
        name = "Action Type", default = "CONNECT",
        items = [("CONNECT", "Connect", "Connect", "", 1)])

    resolution : IntProperty(
        name = 'Resolution', default = 2,
        min = 2, max = 4)

    catenary_a_inv : FloatProperty(
        name = "a_inv", default = 1,
        min = 0.2, max = 2000)

    com_spiral : BoolProperty(
        name = "Use Spiral", default = False)

    com_box : BoolProperty(
        name = "Set Box", default = False)

    use_collection : BoolProperty(
        name = "use_collection", default = True)

    use_material : BoolProperty(
        name = "use_material", default = True)
    
    ClosureDict = {"TYPEA":"Closure.A",
                   "TYPEB":"Closure.B"}
    

    WireNumDict = {'ArmHV.A': 3, 'ArmHV.B': 3, 'ArmHV.C': 3,
                   'ArmHV.D': 3, 'ArmHV.E': 3, 'ArmHV.F': 6,
                   'ArmLV.A': 3, 'ArmLV.B': 3, 'ArmLV.C': 3, 'ArmLV.D': 3, 'ArmLV.E':1, 
                   'ArmCom.A': 2, 'ArmCom.B': 1, 'ArmCom.C': 2, 'ArmCom.D': 3,
                   'ArmWire.A': 1,
                   'Cap.A' : 1,
                   }

    RadiusDict = {'GW':0.005, 'HV':0.01, 'LV':0.01, 'Com':0.01, "WIRE":0.005}
    
    @classmethod
    def poll(cls, context):

        obj = context.object

        if (obj == None):
            return False
        if (obj.select_get() == False):
            return False
        
        if (not obj.poleprops.pole_type in ["TELEGRAPH", "TELEGRAPH_ARM"]):
            return False
        
        return True

    def draw(self, context):

        layout = self.layout

        layout.prop(self, "action")
        layout.separator()
        layout.prop(self, "catenary_a_inv")

        row = layout.row()
        row.prop(self, "com_spiral")
        row.prop(self, "com_box")
        
        layout.prop(self, "resolution")

    def extract_co_listB(self, co_list, base_0, base_1, num):
        #相手に近いほうを抽出
        
        def sort_key(zipped):
            co_xy     = zipped[0].copy(); co_xy.z = 0
            base_0_xy = zipped[1].copy(); base_0_xy.z = 0
            base_1_xy = zipped[2].copy(); base_1_xy.z = 0
            return (co_xy-base_0_xy).dot(base_1_xy-base_0_xy)

        zipped = list(zip(co_list, [base_0]*len(co_list), [base_1]*len(co_list)))
        co_list_sorted = sorted(zipped, key=sort_key, reverse=True)

        #print("num", num)
        #print("co_list", co_list)
        #print("sorted", list(zip(*co_list_sorted)))
        
        co_list = list(zip(*co_list_sorted))[0][0:num]

        return co_list

    
    def sort_co_list(self, co_list0, co_list1, base_0, base_1):
        #ポールとの位置関係でソート

        def sort_key(zipped):
            co   = zipped[0].copy(); 
            base = zipped[1].copy();
            return (co-base).length

        zipped_0 = list(zip(co_list0, [base_0]*len(co_list0)))
        zipped_1 = list(zip(co_list1, [base_1]*len(co_list0)))

        co_list0_sorted = sorted(zipped_0, key=sort_key)
        co_list1_sorted = sorted(zipped_1, key=sort_key)

        co_list0 = list(zip(*co_list0_sorted))[0]
        co_list1 = list(zip(*co_list1_sorted))[0]

        return co_list0, co_list1


    def get_co_list_list(self, tgt_list_list, poles, depsgraph):
        
        co_list_list = [[], []]
        for index, (tgt_list, co_list) in enumerate(zip(tgt_list_list, co_list_list)):

            for tgt in tgt_list:
                co_list_for_each = []
                mesh = tgt.evaluated_get(depsgraph).to_mesh()
                verts = mesh.vertices
                for v in verts:
                    for g in v.groups:
                        
                        if g.group == 0 and g.weight == 1:
                            co_world = tgt.matrix_world @ v.co
                            co_list_for_each.append(co_world)

                # blender3.3以降は Attribute に移行した
                if len(co_list_for_each) == 0:
                    g_name = tgt.vertex_groups[0].name
                    attrib_data = mesh.attributes[g_name].data
                    for v_index, v in enumerate(verts):
                        if (attrib_data[v_index].value == 1):
                            co_world = tgt.matrix_world @ v.co
                            co_list_for_each.append(co_world)
                            
                m = re.match("\w*.[A-Z]", tgt.data.name)
                name_base = m.group()
                num = self.WireNumDict[name_base]

                #print("***", tgt.data.name, num)
                #print(co_list_for_each)
                    
                if (index == 0):
                    co_list_for_each = self.extract_co_listB(co_list_for_each, poles[0].location, poles[1].location, num)
                if (index == 1):
                    co_list_for_each = self.extract_co_listB(co_list_for_each, poles[1].location, poles[0].location, num)

                co_list += co_list_for_each

        return co_list_list

    def set_curve(self, context, co_list0, co_list1, name, use_spiral = False, set_box = False, radius = 0.01):
        curveData = self.make_curve(name)

        for co_0, co_1 in zip(co_list0, co_list1):
            points = curveData.splines.new('POLY').points
            points.add(1)
        
            points[0].co = Vector((co_0[0], co_0[1], co_0[2], 1))
            points[1].co = Vector((co_1[0], co_1[1], co_1[2], 1))
            points[0].radius = radius
            points[1].radius = radius
            
        material = None
        if (context.preferences.addons[addon_name].preferences.use_material == True):
            material = self.load_material(".Pole.Wire")
            
        curveOB = self.make_curve_object(curveData.name, curveData, material, context)
        mod_cat = self.add_GN_modifier(curveOB, '.GN.Pole.Catenary')
        curveOB.poleprops.pole_type = "CATENARY"
        curveOB.poleprops.pole_subtype = name
            
        a_inv = self.catenary_a_inv
        paramList = [("Tension", 20/a_inv),
                     ("Randomness", 0)]

        length = (co_0-co_1).length
        if use_spiral == True:
            paramList.append(("Count"     , int(length * 16)))
            paramList.append(("SpiralTilt", int(length * 2) * 2 * 3.14159))

        self.set_GN_inputs(mod_cat, paramList)
                
            
        if (set_box == True):
            m_name = self.ClosureDict["TYPEA"]
            col = self.get_asset_collection(context, "PoleParts", context.scene.collection)
            boxOB = self.load_object(context, m_name, col)
                
            mod_close = self.add_GN_modifier(curveOB, '.GN.SetAlongCurve', 'ClosureBox')
            paramList = [("Object", boxOB),
                         ("Factor", 0.05)]
            self.set_GN_inputs(mod_close, paramList)

        if (use_spiral == True):
            mod_mesh = self.add_GN_modifier(curveOB, '.GN.Pole.CurveToMeshDual')
        else:
            mod_mesh = self.add_GN_modifier(curveOB, '.GN.Pole.CurveToMesh')

        paramList = [("Material", material),
                     ("Resolution", 8) ]
        self.set_GN_inputs(mod_mesh, paramList)

    def execute_arm(self, context, object0):

        for obj in context.selected_objects:
            if obj != object0:
                object1 = obj

        if (object1.poleprops.pole_type != "TELEGRAPH_ARM"):
            self.report({"WARNING"},
                        "Not telegraph arm pair")
            return {'FINISHED'}

        if (object0.poleprops.pole_subtype != object1.poleprops.pole_subtype):
            self.report({"WARNING"},
                        "Not correct pair type")
            return {'FINISHED'}

        wire_type = object0.poleprops.pole_subtype

        poles = [object0.parent, object1.parent]

        tgt = [[object0], [object1]]

        depsgraph = bpy.context.evaluated_depsgraph_get()
        co_list_list = self.get_co_list_list(tgt, poles, depsgraph)

        use_spiral = False
        use_box = False
        if (wire_type == "Com"):
            use_spiral = self.com_spiral
            use_box    = self.com_box

        radius = self.RadiusDict[wire_type]

        if (co_list_list[0] != [] and co_list_list[1] != []):
            co_list_list[0], co_list_list[1] = self.sort_co_list(co_list_list[0], co_list_list[1], poles[0].location, poles[1].location)
            self.set_curve(context, co_list_list[0], co_list_list[1], 'ElectricalWire'+wire_type, use_spiral, use_box, radius=radius)
            
        return {'FINISHED'}

    def execute(self, context):

        if (len(bpy.context.selected_objects) != 2):
            self.report({"WARNING"},
                        "Catenary could not be completed. Operation Cancelled")
            return {'CANCELLED'}


        object = context.view_layer.objects.active
        if (object.poleprops.pole_type == "TELEGRAPH_ARM"):
            return self.execute_arm(context, object)
        
        poles = context.selected_objects

        children_list = [[], []]
        for o in bpy.data.objects:
            if o.parent == poles[0]:
                children_list[0].append(o)
            if o.parent == poles[1]:
                children_list[1].append(o)

        num_LV = 0
        num_Com = 0
        GW_tgt = []
        HV_tgt = []
        LV_tgt = []
        Com_tgt = []
        for children in children_list:
            GW_list = []
            HV_list = []
            LV_list = []
            Com_list = []
            for child in children:
                parts_name = child.data.name
                if re.match("Cap.A", parts_name):
                    GW_list.append(child)
                if re.match("ArmHV.*", parts_name):
                    HV_list.append(child)
                if re.match("ArmLV.*", parts_name):
                    LV_list.append(child)
                    num_LV += 3
                if re.match("ArmCom.*", parts_name):
                    Com_list.append(child)
                    
                    num_Com += self.WireNumDict[parts_name[0:8]]

            GW_tgt.append(GW_list)
            HV_tgt.append(HV_list)
            LV_tgt.append(LV_list)
            Com_tgt.append(Com_list)

        # [[pole0側のlist], [pole1側のlist]] ってなってる

        num_LV  = int(num_LV  / 2)
        num_Com = int(num_Com / 2)

        depsgraph = bpy.context.evaluated_depsgraph_get()

        co_list_list_GW  = self.get_co_list_list(GW_tgt, poles, depsgraph)
        co_list_list_HV  = self.get_co_list_list(HV_tgt, poles, depsgraph)
        co_list_list_LV  = self.get_co_list_list(LV_tgt, poles, depsgraph)
        co_list_list_Com = self.get_co_list_list(Com_tgt, poles, depsgraph)

        co_list_list = co_list_list_GW
        if (co_list_list[0] != [] and co_list_list[1] != []):
            co_list_list[0], co_list_list[1] = self.sort_co_list(co_list_list[0], co_list_list[1], HV_tgt[0][0].parent.location, HV_tgt[1][0].parent.location)
            self.set_curve(context, co_list_list[0], co_list_list[1], 'ElectricalWireGW', radius=0.005)
            
        co_list_list = co_list_list_HV
        if (co_list_list[0] != [] and co_list_list[1] != []):
            co_list_list[0], co_list_list[1] = self.sort_co_list(co_list_list[0], co_list_list[1], HV_tgt[0][0].parent.location, HV_tgt[1][0].parent.location)
            self.set_curve(context, co_list_list[0], co_list_list[1], 'ElectricalWireHV', radius=0.01)
            
        co_list_list = co_list_list_LV
        if (co_list_list[0] != [] and co_list_list[1] != []):
            co_list_list[0], co_list_list[1] = self.sort_co_list(co_list_list[0], co_list_list[1], LV_tgt[0][0].parent.location, LV_tgt[1][0].parent.location)
            self.set_curve(context, co_list_list[0], co_list_list[1], 'ElectricalWireLV', radius=0.01)

        co_list_list = co_list_list_Com
        if (co_list_list[0] != [] and co_list_list[1] != []):
            co_list_list[0], co_list_list[1] = self.sort_co_list(co_list_list[0], co_list_list[1], Com_tgt[0][0].parent.location, Com_tgt[1][0].parent.location)
            self.set_curve(context, co_list_list[0], co_list_list[1], 'ElectricalWireCom', self.com_spiral, self.com_box, radius=0.01)

        bpy.context.view_layer.objects.active = object

        return {'FINISHED'}


class POLEMAKEROBJ_OT_JoinWires(bpy.types.Operator):
    
    bl_idname = "polemaker.join_wires"
    bl_label = "Catenary Connect"
    bl_description = "Description"
    bl_options = {'REGISTER', 'UNDO'}

    action : StringProperty(
        name="action")

    exclude : BoolProperty(
        name="Exclude Wires with Appendix", default = True,
        description="Exclude wires with additional modifiers")

    def draw(self, context):

        layout = self.layout
        layout.prop(self, "exclude")

    @classmethod
    def poll(cls, context):

        obj = context.object

        if (obj == None):
            return False
        if (obj.select_get() == False):
            return False

        pole_type = obj.poleprops.pole_type
        pole_subtype = obj.poleprops.pole_subtype

        if (pole_type in ['CATENARY']):
            return True

        return False
        

    def execute(self, context):

        obj = context.object
        pole_type = obj.poleprops.pole_type
        pole_subtype = obj.poleprops.pole_subtype

        name = obj.name

        if re.search(r"\.\d\d\d$", name):
            name_candid = re.sub(r"\.\d\d\d$", "", name)

            if (name_candid in context.scene.objects):
                name = name_candid

            context.view_layer.objects.active = bpy.data.objects[name]

        for tgt in context.scene.objects:
            tgt.select_set(False)
            
            if tgt.poleprops.pole_type != pole_type:
                continue
            if tgt.poleprops.pole_subtype != pole_subtype:
                continue

            if (self.exclude == True):
                exclude = False
                for mod in tgt.modifiers:
                    if mod.type != "NODES":
                        exclude = True
                        continue
                    if not mod.node_group.name in ['.GN.Pole.Catenary', '.GN.Pole.CurveToMesh', '.GN.Pole.CurveToMeshDual']:
                        exclude = True
                        continue
                if (exclude):
                    continue
                    
            tgt.select_set(True)

        if context.view_layer.objects.active.select_get() == True:
            bpy.ops.object.join()
        else:
            self.report({'INFO'}, "target wire is not valid")

        return {'FINISHED'}
