
import bpy
import re
import os

from os import path
from mathutils import Vector

dirname = path.dirname(__file__)
addon_name = __package__
pole_name = "PoleMaker"

signal_file = "SignalMark.blend"

class MeshSettingModule:

    def get_asset_collection(self, context, collection_name, parent_col):
        
        if collection_name in bpy.data.collections:
            return bpy.data.collections[collection_name]

        col = bpy.data.collections.new(collection_name)
        parent_col.children.link(col)

        if (col.name in context.view_layer.layer_collection.children):
            context.view_layer.layer_collection.children[col.name].hide_viewport = True
        col.hide_render = True

        return col

    def clean_up_iterator_item(self, src_iterators, item, attrname):

        name = getattr(item, attrname).name

        if re.search(r"\.\d\d\d$", name):
            name_candid = re.sub(r"\.\d\d\d$", "", name)

            
            
            if (name_candid in src_iterators):
                candid = src_iterators[name_candid]

                setattr(item, attrname, candid)

                if (src_iterators[name].users == 0):
                    src_iterators.remove(src_iterators[name])

    def clean_up_material_texture(self, material):

        for node in material.node_tree.nodes:
            if node.type != "TEX_IMAGE":
                continue

            self.clean_up_iterator_item(bpy.data.images, node, "image")

               
    def clean_up_nodegroup(self, node_group):

        self.clean_up_input_material(node_group)

        for node in node_group.nodes:

            if node.type != 'GROUP':
                continue

            node_tree = node.node_tree            
            self.clean_up_nodegroup(node_tree)

            self.clean_up_iterator_item(bpy.data.node_groups, node, "node_tree")
            

    def clean_up_input_material_each(self, input):

        if (input.default_value == None):
            return

        self.clean_up_iterator_item(bpy.data.materials, input, "default_value")
        

    def clean_up_input_material(self, node_group):

        for node in node_group.nodes:

            if not node.type in ['GROUP', 'REPLACE_MATERIAL', 'SET_MATERIAL', 'SWITCH']:
                continue

            for input in node.inputs:
                if (input.type != 'MATERIAL'):
                    continue

                self.clean_up_input_material_each(input)

            if (node.type == 'GROUP'):
                if bpy.app.version >= (4,0,0):
                    for item in node.node_tree.interface.items_tree:
                        if item.in_out == 'INPUT':
                            if (item.socket_type != 'NodeSocketMaterial'):
                                continue

                            self.clean_up_input_material_each(item)
                else:
                    for input in node.node_tree.inputs:
                        if (input.type != 'MATERIAL'):
                            continue
                        
                        self.clean_up_input_material_each(input)

    def make_load_list(self, itemname_list, typename):

        item_to_read = []
        for itemname in itemname_list:
            if not itemname in getattr(bpy.data, typename):
                item_to_read.append(itemname)

        return item_to_read

    def load_NodeGroup(self, node_name):

        if node_name in bpy.data.node_groups:
            return bpy.data.node_groups[node_name]

        #print("Load", node_name)

        pre_material_list = [mat for mat in bpy.data.materials]

        libfile = dirname + os.sep + pole_name + ".blend"
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.node_groups = [node_name]

        node_group = bpy.data.node_groups[node_name]
        
        post_material_list = [mat for mat in bpy.data.materials]

        diff = set(pre_material_list) ^ set(post_material_list)
        for mat in list(diff):
            self.clean_up_material_texture(mat)
        
        self.clean_up_nodegroup(node_group)
        

        return node_group

    def load_NodeGroup_list(self, nodegroup_list):

        nodegroup_to_read = []
        for nodegroup_name in nodegroup_list:
            if not nodegroup_name in bpy.data.node_groups:
                nodegroup_to_read.append(nodegroup_name)

        if len(nodegroup_to_read) == 0:
            return

        pre_material_list = [mat for mat in bpy.data.materials]

        libfile = dirname + os.sep + pole_name + ".blend"
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.node_groups = nodegroup_to_read

        post_material_list = [mat for mat in bpy.data.materials]

        diff = set(pre_material_list) ^ set(post_material_list)

        #print("diff=", diff)

        
        for mat in list(diff):
            self.clean_up_material_texture(mat)


        for nodegroup_name in nodegroup_list:
            nodegroup = bpy.data.node_groups[nodegroup_name]
            self.clean_up_nodegroup(nodegroup)
            
    
    def load_GP(self, GPname):

        if not GPname:
            return

        if (GPname in bpy.data.grease_pencils):
            return

        dirname = path.dirname(__file__)
        libfile = dirname + os.sep + signal_file
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.grease_pencils = [GPname]

    def make_GP(self, context, GPname, use_collection):

        if (GPname == ""):
            print("No GP Name")
            return None

        obj = bpy.data.objects.new(GPname, bpy.data.grease_pencils[GPname])

        col_to_belong = self.get_col_to_belong(context, use_collection)
        col_to_belong.objects.link(obj)
        
        return obj

    def get_col_to_belong(self, context, use_collection):

        obj = context.object

        #col_to_belong = context.scene.collection
        col_to_belong = context.collection
        if (use_collection):
            for col in bpy.data.collections:
                for o in col.objects:
                    if o == obj:
                        col_to_belong = col

        return col_to_belong

    def get_material_from_meshname(self, meshname):

        if (re.match("^[CO][HV]*", meshname)):
            return ".Pole.White"
        if (re.match("Mirror*", meshname)):
            return ".Pole.Orange"
        
        return ".Pole.Basic"

    def simplify_material(self, mesh):
        
        for index, mat in enumerate(mesh.materials):
            if mat == None:
                continue
            name = mat.name 
        
            if re.search(r"\.\d\d\d$", name):
                name_candid = re.sub(r"\.\d\d\d", "", name)
                
                if (name_candid in bpy.data.materials):
                    mat_orig = mat
                    mesh.materials[index] = bpy.data.materials[name_candid]

                    if (mat_orig.users == 0):
                        bpy.data.materials.remove(mat_orig)

    def load_mesh(self, meshname):

        if not meshname:
            return

        if (meshname in bpy.data.meshes):
            return

        #print("load_mesh", meshname)

        libfile = dirname + os.sep + pole_name + ".blend"
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.meshes = [meshname]

        material = None

        if (bpy.context.preferences.addons[addon_name].preferences.use_material == True):
            material_name = self.get_material_from_meshname(meshname)
            material = self.load_material(material_name)

        mesh = bpy.data.meshes[meshname]
        if len(mesh.materials) == 0:
            mesh.materials.append(material)

        index = 0
        for index, mat in enumerate(mesh.materials):
            if mat == None:
                mesh.materials[index] = material

        self.simplify_material(mesh)

        
    def load_mesh_list(self, meshname_list):

        mesh_to_read = self.make_load_list(meshname_list, "meshes")
        
        if len(mesh_to_read) == 0:
            return

        libfile = dirname + os.sep + pole_name + ".blend"
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.meshes = mesh_to_read

        for mesh in mesh_to_read:
            if mesh == None:
                continue

            material = None
            if (bpy.context.preferences.addons[addon_name].preferences.use_material == True):
                material_name = self.get_material_from_meshname(mesh.name)
                material = self.load_material(material_name)
                
            if len(mesh.materials) == 0:
                mesh.materials.append(material)

            index = 0
            for mat in mesh.materials:
                if mat == None:
                    mesh.materials[index] = material
                    
                index += 1

    def load_curve(self, curvename):

        if not curvename:
            return

        if (not curvename in bpy.data.curves):
            dirname = path.dirname(__file__)
            addon_name = __package__
            libfile = dirname + os.sep + pole_name + ".blend"
            with bpy.data.libraries.load(libfile) as (data_from, data_to):
                data_to.curves = [curvename]

    def load_object(self, context, dataname, col_to_belong = None, use_curve = False):

        if (dataname in bpy.data.objects):
            return bpy.data.objects[dataname]
        
        if (dataname == ""):
            print("No Mesh Name")
            return None

        if (use_curve == True):
            self.load_curve(dataname)
            obj = bpy.data.objects.new(dataname, bpy.data.curves[dataname])
        else:
            self.load_mesh(dataname)
            obj = bpy.data.objects.new(dataname, bpy.data.meshes[dataname])

        if (context.preferences.addons[addon_name].preferences.use_material == True):
            for slot in obj.material_slots:
                material = slot.material
                slot.link = "OBJECT"
                slot.material = material
                slot.link = "DATA"

        if (col_to_belong):
            col_to_belong.objects.link(obj)
        else:
            context.scene.collection.objects.link(obj)

        return obj

    def load_material(self, matname):
        if (matname in bpy.data.materials):
            return bpy.data.materials[matname]

        if (matname == ""):
            print("No Material Name")
            return None

        print("load_material", matname)

        libfile = dirname + os.sep + pole_name + ".blend"

        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.materials = [matname]

        return bpy.data.materials[matname]

    def load_material_list(self, materialname_list):

        material_to_read = self.make_load_list(materialname_list, "materials")

        if len(material_to_read) == 0:
            return
        
        libfile = dirname + os.sep + pole_name + ".blend"
        with bpy.data.libraries.load(libfile) as (data_from, data_to):
            data_to.materials = material_to_read
        
    
    def get_radius_at(self, pos_z, telegraph_type):

        top, bottom, btm_radius, top_radius = self.TelegraphDict[telegraph_type]

        radius = ((top-pos_z)*btm_radius + (pos_z-bottom)*top_radius) / (top-bottom)

        #print(radius)
        
        return radius

    def make_on_pole_instance(self, context, pole, objname, pos_z):

        self.load_mesh("EmptyMesh")

        obj = bpy.data.objects.new(objname, bpy.data.meshes["EmptyMesh"])
        
        use_collection = False
        if hasattr(self, "use_collection"):
            use_collection = self.use_collection

        col_to_belong = self.get_col_to_belong(context, use_collection)
        col_to_belong.objects.link(obj)

        obj.location.z = pos_z

        obj.parent = pole

        return obj
        
    
    def make_on_pole(self, context, pole, meshname, pos_z, offset = 0, radius = 0):
    
        if (meshname == ""):
            print("No Mesh Name")
            return None

        self.load_mesh(meshname)
        
        obj = bpy.data.objects.new(meshname, bpy.data.meshes[meshname])
        
        if (context.preferences.addons[addon_name].preferences.use_material == True):
            for slot in obj.material_slots:
                material = slot.material
                slot.link = "OBJECT"
                slot.material = material
                slot.link = "DATA"

        use_collection = False
        if hasattr(self, "use_collection"):
            use_collection = self.use_collection

        col_to_belong = self.get_col_to_belong(context, use_collection)
        col_to_belong.objects.link(obj)

        if (radius == 0):
            radius = self.get_radius_at(pos_z, self.telegraph_type)
        
        offset_x = radius * offset

        obj.location = Vector((offset_x, 0, pos_z))
        
        obj.parent = pole

        return obj

    def make_obj(self, context, meshname):

        if (meshname == ""):
            print("No Mesh Name")
            return None

        self.load_mesh(meshname)

        obj = bpy.data.objects.new(meshname, bpy.data.meshes[meshname])
    
        if (context.preferences.addons[addon_name].preferences.use_material == True):
            for slot in obj.material_slots:
                material = slot.material
                slot.link = "OBJECT"
                slot.material = material
                slot.link = "DATA"
                
        use_collection = False
        if hasattr(self, "use_collection"):
            use_collection = self.use_collection

        col_to_belong = self.get_col_to_belong(context, use_collection)
        col_to_belong.objects.link(obj)

        return obj


